import json
from copy import copy
from io import BytesIO
from json.decoder import JSONDecodeError
from typing import Any, Dict, ItemsView, KeysView, Optional, ValuesView

import yaml

from .http_multipart_message_parser import parse_multipart_message
from .http_query_string import parse_qs


class HttpMessage(str):
    pass


class CompositeHttpMessage(HttpMessage):
    def __init__(self, data: dict) -> None:
        self.data = data

    def get(self, name: str, default: Optional[Any] = None) -> Any:
        if name in self:
            return self[name]

        return default

    def __getitem__(self, key) -> Any:
        return self.data[key]

    def __contains__(self, key):
        return key in self.data

    def __copy__(self):
        return type(self)(copy(self.data))

    def __iter__(self):
        return iter(self.data)

    def items(self) -> ItemsView:
        return self.data.items()

    def values(self) -> ValuesView:
        return self.data.values()

    def keys(self) -> KeysView:
        return self.data.keys()


class YamlHttpMessage(CompositeHttpMessage):
    @staticmethod
    def from_bytes(body: BytesIO, encoding: str = "utf8") -> "YamlHttpMessage":
        body.seek(0)
        decoded_input = body.read().decode(encoding)

        parsed_body: Dict[str, Any] = {}
        try:
            parsed_body = yaml.safe_load_all(decoded_input)  # type: ignore
        except JSONDecodeError:
            ...  # ignore

        return YamlHttpMessage(parsed_body)


class FormHttpMessage(CompositeHttpMessage):
    @staticmethod
    def from_bytes(body: BytesIO, encoding: str = "utf8") -> "FormHttpMessage":
        body.seek(0)
        decoded_input = body.read().decode(encoding)
        fields = parse_qs(decoded_input)
        return FormHttpMessage(fields)


class JsonHttpMessage(CompositeHttpMessage):
    @staticmethod
    def from_bytes(body: BytesIO, encoding: str = "utf8") -> "JsonHttpMessage":
        body.seek(0)
        decoded_input = body.read().decode(encoding)

        parsed_body: Dict[str, Any] = {}
        try:
            parsed_body = json.loads(decoded_input)
        except JSONDecodeError:
            ...  # ignore

        return JsonHttpMessage(parsed_body)


class MultipartHttpMessage(CompositeHttpMessage):
    @staticmethod
    def from_bytes(body: BytesIO, boundary: str, encoding: str = "utf8") -> "MultipartHttpMessage":
        body.seek(0)
        fields = parse_multipart_message(body.read(), boundary, encoding)

        return MultipartHttpMessage(fields)


__all__ = [
    "HttpMessage",
    "CompositeHttpMessage",
    "FormHttpMessage",
    "JsonHttpMessage",
    "MultipartHttpMessage",
    "YamlHttpMessage",
]



import base64
import re
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from ipaddress import AddressValueError, IPv4Address, IPv6Address
from typing import Any, Pattern, Union
from uuid import UUID

from .errors import FormatValidationError
from .iso_datetime import (
    parse_iso_date_string,
    parse_iso_datetime_string,
    parse_iso_duration_string,
    parse_iso_time_string,
)


def validate_format_pattern(value: Any) -> Pattern[str]:
    try:
        return re.compile(value)
    except Exception:
        raise FormatValidationError(expected_format="pattern")


def validate_format_bytes(value: Any) -> bytes:
    try:
        return base64.b64decode(value)
    except Exception:
        pass

    raise FormatValidationError(expected_format="byte")


FALSY_EXPRESSION = {0, "0", "no", "n", "nope", "false", "f", "off"}
TRUTHY_EXPRESSION = {1, "1", "ok", "yes", "y", "yup", "true", "t", "on"}


def validate_format_boolean(value: Any) -> bool:
    if value in FALSY_EXPRESSION:
        return False

    if value in TRUTHY_EXPRESSION:
        return True

    raise FormatValidationError(expected_format="boolean")


def validate_format_datetime(value: Any) -> datetime:
    if isinstance(value, datetime):
        return value

    value = str(value)
    try:
        return parse_iso_datetime_string(value)
    except ValueError:
        raise FormatValidationError(expected_format="date-time")


def validate_format_date(value: Any) -> date:
    if isinstance(value, date):
        return value

    value = str(value)
    try:
        return parse_iso_date_string(value)
    except ValueError:
        raise FormatValidationError(expected_format="date")


def validate_format_time(value: Any) -> time:
    if isinstance(value, time):
        return value
    value = str(value)
    try:
        return parse_iso_time_string(value)
    except ValueError:
        raise FormatValidationError(expected_format="time")


def validate_format_time_duration(value: Any) -> timedelta:
    if isinstance(value, timedelta):
        return value

    value = str(value)
    try:
        return parse_iso_duration_string(value)
    except Exception:
        raise FormatValidationError(expected_format="time-duration")


# https://www.w3.org/TR/html5/forms.html#valid-e-mail-address
EMAIL_REGEX = re.compile(
    r"^[a-zA-Z0-9.!#$%&'*+\/=?^_`{|}~-]+"
    r"@[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?(?:\.[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)*$",
    re.I,
)


def validate_format_email(value: str) -> str:
    """
    Keep in mind this validator willfully violates RFC 5322, the best way to invalidate email address is to send
    a message and receive confirmation from the recipient.
    """
    if not EMAIL_REGEX.match(value):
        raise FormatValidationError(expected_format="email")
    if ".." in value:
        raise FormatValidationError(expected_format="email")

    return value


HOSTNAME_REGEX = re.compile(r"^[a-z0-9](?:[a-z0-9-]{0,61}[a-z0-9])?(?:\.[a-z0-9](?:[-0-9a-z]{0,61}[0-9a-z])?)*$", re.I,)


def validate_format_hostname(value: str) -> str:
    if not HOSTNAME_REGEX.match(value):
        raise FormatValidationError(expected_format="hostname")

    return value


def validate_format_decimal(value: Any) -> Decimal:
    if isinstance(value, Decimal):
        return value

    try:
        value = Decimal(value)
        if not value.is_finite():
            raise FormatValidationError(expected_format="decimal")

    except Exception:
        raise FormatValidationError(expected_format="decimal")

    return value


def validate_format_ip_address_v4(value: Any) -> IPv4Address:
    try:
        return IPv4Address(value)
    except AddressValueError:
        raise FormatValidationError(expected_format="ip-address-v4")


def validate_format_ip_address_v6(value: Any) -> IPv6Address:
    try:
        return IPv6Address(value)
    except AddressValueError:
        raise FormatValidationError(expected_format="ip-address-v6")


def validate_format_ip_address(value: Any) -> Union[IPv4Address, IPv6Address]:
    try:
        return IPv4Address(value)
    except AddressValueError:
        try:
            return IPv6Address(value)
        except AddressValueError:
            raise FormatValidationError(expected_format="ip-address")


SEMVER_REGEX = re.compile(
    r"^((([0-9]+)\.([0-9]+)\.([0-9]+)(?:-([0-9a-z-]+(?:\.[0-9a-z-]+)*))?)(?:\+([0-9a-z-]+(?:\.[0-9a-z-]+)*))?)$", re.I,
)


def validate_format_semver(value: Any) -> str:
    value = str(value)
    if not SEMVER_REGEX.match(value):
        raise FormatValidationError(expected_format="semver")

    return value


URI_REGEX = re.compile(r"^(?:[a-z][a-z0-9+-.]*:)(?:\\/?\\/)?[^\s]*$", re.I)


def validate_format_uri(value: Any) -> str:
    value = str(value)
    if not URI_REGEX.match(value):
        raise FormatValidationError(expected_format="uri")

    return value


URL_REGEX = re.compile(
    r"^(?:(?:https?|ftp):\/\/)(?:\S+(?::\S*)?@)?(?:(?!10(?:\.\d{1,3}){3})(?!127(?:\.\d{1,3}){3})(?!169\.254(?:\.\d{1,3}){2})(?!192\.168(?:\.\d{1,3}){2})(?!172\.(?:1[6-9]|2\d|3[0-1])(?:\.\d{1,3}){2})(?:[1-9]\d?|1\d\d|2[01]\d|22[0-3])(?:\.(?:1?\d{1,2}|2[0-4]\d|25[0-5])){2}(?:\.(?:[1-9]\d?|1\d\d|2[0-4]\d|25[0-4]))|(?:(?:[a-z\u00a1-\uffff0-9_]+-?)*[a-z\u00a1-\uffff0-9_]+)(?:\.(?:[a-z\u00a1-\uffff0-9_]+-?)*[a-z\u00a1-\uffff0-9_]+)*(?:\.(?:[a-z\u00a1-\uffff]{2,})))(?::\d{2,5})?(?:\/[^\s]*)?$",
    re.I | re.U,
)


def validate_format_url(value: Any) -> str:
    value = str(value)
    if not URL_REGEX.match(value):
        raise FormatValidationError(expected_format="url")

    return value


def validate_format_uuid(value: Any) -> UUID:
    try:
        return UUID(value)
    except Exception:
        raise FormatValidationError(expected_format="uuid")


__all__ = [
    "validate_format_boolean",
    "validate_format_bytes",
    "validate_format_date",
    "validate_format_datetime",
    "validate_format_decimal",
    "validate_format_email",
    "validate_format_hostname",
    "validate_format_ip_address",
    "validate_format_ip_address_v4",
    "validate_format_ip_address_v6",
    "validate_format_pattern",
    "validate_format_semver",
    "validate_format_time",
    "validate_format_time_duration",
    "validate_format_uri",
    "validate_format_url",
    "validate_format_uuid",
]



import os
from copy import copy
from typing import Any

from chocs.http_request import HttpRequest
from chocs.http_response import HttpResponse
from chocs.middleware.middleware import MiddlewareHandler, MiddlewarePipeline
from chocs.routing import Route
from chocs.types import HttpHandlerFunction


class ServerlessFunction:
    def __init__(
        self,
        function: HttpHandlerFunction,
        route: Route = Route("/"),
        middleware_pipeline: MiddlewarePipeline = MiddlewarePipeline(),
    ):
        self._function = function
        self._route = route
        self._middleware_pipeline = MiddlewarePipeline(middleware_pipeline.queue)
        self._middleware_enabled = False

        def _function_middleware(_request: HttpRequest, _next: MiddlewareHandler) -> HttpResponse:
            _route = copy(route)
            route._parameters = _request.path_parameters
            _request.route = _route

            return function(_request)

        self.middleware_pipeline.append(_function_middleware)

    @property
    def function(self) -> HttpHandlerFunction:
        return self._function

    @property
    def route(self) -> Route:
        return self._route

    @property
    def middleware_pipeline(self) -> MiddlewarePipeline:
        return self._middleware_pipeline

    @property
    def middleware_enabled(self) -> bool:
        return self._middleware_enabled

    @middleware_enabled.setter
    def middleware_enabled(self, value: bool) -> None:
        self._middleware_enabled = value

    def __call__(self, *args) -> Any:
        request = args[0]
        route = copy(self.route)
        route._parameters = request.path_parameters
        request.route = route
        request.attributes["__handler__"] = self._function
        if self._middleware_enabled and not self._middleware_pipeline.empty:
            return self._middleware_pipeline(*args)

        return self._function(*args)


IS_AWS_ENVIRONMENT = bool(
    os.environ.get("AWS_LAMBDA_FUNCTION_VERSION")
    or os.environ.get("LAMBDA_RUNTIME_DIR")
    or os.environ.get("AWS_LAMBDA_FUNCTION_NAME")
)


__all__ = ["ServerlessFunction", "IS_AWS_ENVIRONMENT"]



from numbers import Number
from typing import Any, Callable, List, Union

from .errors import EnumValidationError, TypeValidationError


def validate_string(value: Any) -> str:
    if isinstance(value, str):
        return value

    raise TypeValidationError(value=value, expected_type=str)


def validate_boolean(value: Any) -> bool:
    if value is True or value is False:
        return value

    raise TypeValidationError(expected_type=bool)


def validate_enum(value: Any, values: List[Union[str, int, float, bool]]) -> Union[str, int, float, bool]:
    for item in values:  # `if value in values` expression does casting and we dont want it
        if isinstance(value, bool):
            if item is value:
                return value
        else:
            if item == value:
                return value

    raise EnumValidationError(expected_values=values)


def validate_integer(value: Any) -> int:
    if isinstance(value, int) and value is not True and value is not False:
        return value

    raise TypeValidationError(expected_type=int)


def validate_number(value: Any) -> Number:
    if value is True or value is False:
        raise TypeValidationError(expected_type=Number)

    if isinstance(value, Number):
        return value

    raise TypeValidationError(expected_type=Number)


def validate_array(value: list) -> list:
    if not isinstance(value, list):
        raise TypeValidationError(expected_type="array")

    return value


def validate_object(value: dict) -> dict:
    if not isinstance(value, dict):
        raise TypeValidationError(expected_type="object")
    return value


def validate_nullable(value: Any, validator: Callable) -> Any:
    if value is None:
        return None

    return validator(value)


__all__ = [
    "validate_array",
    "validate_boolean",
    "validate_enum",
    "validate_integer",
    "validate_number",
    "validate_nullable",
    "validate_object",
    "validate_string",
]



from typing import Generator, KeysView, Optional, Sequence, Union, ValuesView


def _normalize_header_name(name: str) -> str:
    """
    According to rfc https://www.ietf.org/rfc/rfc2616.txt header names are case insensitive,
    thus they can be normalized to ease usage of Headers class.
    :param name:
    :return:
    """
    name = name.lower()
    if name.startswith("http_"):
        name = name[5:]

    return name.replace("_", "-")


def _normalize_headers(headers: dict) -> dict:
    normalized = {}
    for name, value in headers.items():
        if isinstance(value, list):
            normalized[_normalize_header_name(name)] = value
        else:
            normalized[_normalize_header_name(name)] = [value]

    return normalized


class HttpHeaders:
    """
    Dict-like object containing http headers. Header names are case-insensitive, and
    their values are internally stored as sequences to conform RFC-2616 standard.
    .. _RFC-2616 Section 4.2: https://www.w3.org/Protocols/rfc2616/rfc2616-sec4.html#sec4.2
    """

    def __init__(self, headers: Optional[dict] = None):

        headers = {} if headers is None else headers
        self._headers = _normalize_headers(headers)

    def set(self, name: str, value: str) -> None:
        """
        Appends new value to the header, if header does not exists it will get created.
        """
        normalized_name = _normalize_header_name(name)
        if normalized_name not in self._headers:
            self._headers[normalized_name] = []

        self._headers[normalized_name].append(value)

    def override(self, name: str, value: Union[str, list]) -> None:
        normalized_name = _normalize_header_name(name)
        self._headers[normalized_name] = value if isinstance(value, list) else [value]

    def get(self, name: str, default: str = "") -> Union[str, Sequence[str]]:
        if name in self:
            return self.__getitem__(name)
        return default

    def __setitem__(self, name: str, value: Sequence[str]) -> None:
        """
        Sets value for header. Value must be valid sequence of strings.
        """
        self._headers[_normalize_header_name(name)] = value

    def __getitem__(self, name: str) -> Union[str, Sequence[str]]:
        """
        Returns string if header is unique otherwise sequence of strings is returned.
        """
        value = self._headers.get(_normalize_header_name(name), None)
        if value is None:
            return ""

        if len(value) == 1:
            return value[0]

        return value

    def __contains__(self, name: str) -> bool:
        return _normalize_header_name(name) in self._headers

    def items(self) -> Generator:
        for key, values in self._headers.items():
            if len(values) == 1:
                yield key, values[0]
                continue
            for value in values:
                yield key, value

    def values(self) -> ValuesView[Union[str, Sequence[str]]]:
        return self._headers.values()

    def keys(self) -> KeysView[str]:
        return self._headers.keys()

    def __eq__(self, other) -> bool:
        if not isinstance(other, HttpHeaders):
            return False

        return self._headers == other._headers


__all__ = ["HttpHeaders"]



from .application_middleware import ApplicationMiddleware
from .middleware import Middleware, MiddlewareCursor, MiddlewareFunction, MiddlewareHandler, MiddlewarePipeline
from .open_api_middleware import OpenApiMiddleware



from typing import Callable, Dict, List, Union

from .errors import (
    AdditionalPropertyError,
    MaximumPropertyError,
    MinimumPropertyError,
    PropertyNameError,
    PropertyValueError,
    RequiredPropertyError,
    ValidationError,
)


def validate_object_properties(
    obj: dict, properties: Dict[str, Callable], additional_properties: Union[bool, Callable] = True,
) -> dict:

    property_length = 0
    for key, value in obj.items():
        property_length += 1
        if key not in properties:
            if not additional_properties:
                raise AdditionalPropertyError(property_name=key)
            if callable(additional_properties):
                try:
                    obj[key] = additional_properties(value)
                except PropertyValueError as error:
                    raise PropertyValueError(
                        property_name=key + "." + error.context["property_name"],
                        validation_error=error.context["validation_error"],
                    )
                except ValidationError as error:
                    raise PropertyValueError(
                        property_name=key, validation_error=str(error), sub_code=error.code,
                    )
                except ValueError as error:
                    raise PropertyValueError(property_name=key, validation_error=str(error))
            continue
        try:
            obj[key] = properties[key](value)
        except PropertyValueError as error:
            raise PropertyValueError(
                sub_code=error.code,
                property_name=key + "." + error.context["property_name"],
                validation_error=error.context["validation_error"],
            )
        except ValidationError as error:
            raise PropertyValueError(property_name=key, validation_error=str(error), sub_code=error.code)
        except ValueError as error:
            raise PropertyValueError(property_name=key, validation_error=str(error))

    return obj


def validate_object_property_names(obj: dict, property_names: Callable) -> dict:
    for name in obj.keys():
        try:
            property_names(name)
        except ValidationError as error:
            raise PropertyNameError(sub_code=error.code, property_name=name, validation_error=str(error))
        except ValueError as error:
            raise PropertyNameError(sub_code="error", property_name=name, validation_error=str(error))

    return obj


def validate_object_required_properties(obj: dict, required_properties: List[str]) -> dict:
    given_properties = obj.keys()
    missing_properties = [key for key in required_properties if key not in given_properties]

    if missing_properties:
        raise RequiredPropertyError(property_name=missing_properties[0])

    return obj


def validate_object_minimum_properties(obj: dict, expected_minimum: int) -> dict:
    if len(obj) < expected_minimum:
        raise MinimumPropertyError(expected_minimum=expected_minimum)
    return obj


def validate_object_maximum_properties(obj: dict, expected_maximum: int) -> dict:
    if len(obj) > expected_maximum:
        raise MaximumPropertyError(expected_maximum=expected_maximum)

    return obj


__all__ = [
    "validate_object_maximum_properties",
    "validate_object_minimum_properties",
    "validate_object_properties",
    "validate_object_property_names",
    "validate_object_required_properties",
]
